@@ -1,4 +1,4 @@
-// Copyright 2015-2022 The NATS Authors
+// Copyright 2015-2024 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -41,6 +41,7 @@
 #include "url.h"
 #include "srvpool.h"
 #include "msg.h"
+#include "dispatch.h"
 #include "asynccb.h"
 #include "hash.h"
 #include "stats.h"
@@ -166,6 +167,7 @@ typedef struct __natsServerInfo
     char        *version;
     bool        authRequired;
     bool        tlsRequired;
+    bool        tlsAvailable;
     int64_t     maxPayload;
     char        **connectURLs;
     int         connectURLsCount;
@@ -223,6 +225,7 @@ struct __natsOptions
     bool                    pedantic;
     bool                    allowReconnect;
     bool                    secure;
+    bool                    tlsHandshakeFirst;
     int                     ioBufSize;
     int                     maxReconnect;
     int64_t                 reconnectWait;
@@ -264,13 +267,23 @@ struct __natsOptions
     int64_t                 pingInterval;
     int                     maxPingsOut;
     int                     maxPendingMsgs;
+    int64_t                 maxPendingBytes;
 
     natsSSLCtx              *sslCtx;
 
     void                    *evLoop;
     natsEvLoopCallbacks     evCbs;
 
-    bool                    libMsgDelivery;
+    // If set to false, the client will start a per-subscription "own"
+    // thread to deliver messages to the user callbacks. If true, a shared
+    // thread out of a thread pool is used. natsClientConfig controls the pool
+    // size.
+    bool                    useSharedDispatcher;
+
+    // If set to false, the client will start a per-connection dedicated thread
+    // to deliver reply messages to the user callbacks. If true, a shared thread
+    // out of a thread pool is used. natsClientConfig controls the pool size.
+    bool                    useSharedReplyDispatcher;
 
     int                     orderIP; // possible values: 0,4,6,46,64
 
@@ -331,26 +344,6 @@ struct __natsOptions
     // Custom message payload padding size
     int payloadPaddingSize;
 };
-
-typedef struct __nats_MsgList
-{
-    natsMsg     *head;
-    natsMsg     *tail;
-    int         msgs;
-    int         bytes;
-
-} nats_MsgList;
-
-typedef struct __natsMsgDlvWorker
-{
-    natsMutex       *lock;
-    natsCondition   *cond;
-    natsThread      *thread;
-    bool            shutdown;
-    nats_MsgList     msgList;
-
-} natsMsgDlvWorker;
-
 typedef struct __pmInfo
 {
     char                *subject;
@@ -379,6 +372,28 @@ struct __jsCtx
     bool                closed;
 };
 
+typedef struct __jsFetch
+{
+    struct jsOptionsPullSubscribeAsync opts;
+
+    natsStatus  status;
+
+    // Stats
+    int64_t     startTimeMillis;
+    int         receivedMsgs;
+    int64_t     receivedBytes;
+    int         deliveredMsgs;
+    int64_t     deliveredBytes;
+    int         requestedMsgs;
+
+    // Timer for the fetch expiration. We leverage the existing jsi->hbTimer for
+    // checking missed heartbeats.
+    natsTimer   *expiresTimer;
+
+    // Matches jsi->fetchID
+    char        replySubject[NATS_DEFAULT_INBOX_PRE_LEN + NUID_BUFFER_LEN + 32]; // big enough for {INBOX}.number
+} jsFetch;
+
 typedef struct __jsSub
 {
     jsCtx               *js;
@@ -392,6 +407,7 @@ typedef struct __jsSub
     bool                dc; // delete JS consumer in Unsub()/Drain()
     bool                ackNone;
     uint64_t            fetchID;
+    jsFetch             *fetch;
 
     // This is ConsumerInfo's Pending+Consumer.Delivered that we get from the
     // add consumer response. Note that some versions of the server gather the
@@ -404,7 +420,6 @@ typedef struct __jsSub
     int64_t             hbi;
     bool                active;
     natsTimer           *hbTimer;
-    natsMsg             *mhMsg;
 
     char                *cmeta;
     uint64_t            sseq;
@@ -494,6 +509,21 @@ struct __kvWatcher
 
 };
 
+typedef struct __natsSubscriptionControlMessages
+{
+    struct
+    {
+        natsMsg *timeout;
+        natsMsg *close;
+        natsMsg *drain;
+    } sub;
+    struct
+    {
+        natsMsg *expired;
+        natsMsg *missedHeartbeat;
+    } fetch;
+} natsSubscriptionControlMessages;
+
 struct __natsSubscription
 {
     natsMutex                   *mu;
@@ -503,22 +533,28 @@ struct __natsSubscription
     // This is non-zero when auto-unsubscribe is used.
     uint64_t                    max;
 
-    // This is updated in the delivery thread (or NextMsg) and indicates
-    // how many message have been presented to the callback (or returned
-    // from NextMsg). Like 'msgs', this is also used to determine if we
-    // have reached the max number of messages.
+    // We always have a dispatcher to keep track of things, even if the
+    // subscription is sync. The dispatcher is set up at the subscription
+    // creation time, and may point to a dedicated thread that uses sub's own
+    // dispatchQueue, or a shared worker with a shared queue, which
+    // dispatcher->queue then points to.
+    natsDispatcher *dispatcher;
+    natsDispatcher ownDispatcher;
+
+    // These are a signals to the sub's async dispatcher thread that something
+    // happened - draining or closing the subscription, or some sort of a
+    // timeout. Since these are optional, we only allocate them when starting an
+    // async dispatcher.
+    natsSubscriptionControlMessages *control;
+
+    // This is updated in the delivery thread (or NextMsg) and indicates how
+    // many message have been presented to the callback (or returned from
+    // NextMsg). Together with the messages pending dispatch in
+    // dispatch->queue, this is also used to determine if we have reached the
+    // max number of messages.
     uint64_t                    delivered;
-
-    // The list of messages waiting to be delivered to the callback (or
-    // returned from NextMsg).
-    nats_MsgList                msgList;
-
-    // True if msgList.count is over pendingMax
+    // True if ownDispatcher.queue.msgs is over pendingMax
     bool                        slowConsumer;
-
-    // Condition variable used to wait for message delivery.
-    natsCondition               *cond;
-
     // The subscriber is closed (or closing).
     bool                        closed;
 
@@ -534,11 +570,7 @@ struct __natsSubscription
     int64_t                     drainTimeout;
     // This is set if the flush failed and will prevent the connection for pushing further messages.
     bool                        drainSkip;
-
-    // Same than draining but for the global delivery situation.
-    // This boolean will be switched off when processed, as opposed
-    // to draining that once set does not get reset.
-    bool                        libDlvDraining;
+    natsCondition               *drainCond;
 
     // If true, the subscription is closed, but because the connection
     // was closed, not because of subscription (auto-)unsubscribe.
@@ -559,13 +591,6 @@ struct __natsSubscription
     // Reference to the connection that created this subscription.
     struct __natsConnection     *conn;
 
-    // Delivery thread (for async subscription).
-    natsThread                  *deliverMsgsThread;
-
-    // If message delivery is done by the library instead, this is the
-    // reference to the worker handling this subscription.
-    natsMsgDlvWorker            *libDlvWorker;
-
     // Message callback and closure (for async subscription).
     natsMsgHandler              msgCb;
     void                        *msgCbClosure;
@@ -588,7 +613,6 @@ struct __natsSubscription
 
     // For JetStream
     jsSub                       *jsi;
-
 };
 
 typedef struct __natsPong
@@ -712,7 +736,7 @@ struct __natsConnection
     // New Request style
     char                respId[NATS_MAX_REQ_ID_LEN+1];
     int                 respIdPos;
-    int                 respIdVal;
+    char                respIdVal;
     char                *respSub;   // The wildcard subject
     natsSubscription    *respMux;   // A single response subscription
     natsStrHash         *respMap;   // Request map for the response msg
@@ -748,62 +772,9 @@ struct __natsConnection
     } srvVersion;
 };
 
-//
-// Library
-//
-
-void
-natsSys_Init(void);
-
-void
-natsLib_Retain(void);
-
-void
-natsLib_Release(void);
-
-int64_t
-nats_setTargetTime(int64_t timeout);
-
-void
-nats_resetTimer(natsTimer *t, int64_t newInterval);
-
-void
-nats_stopTimer(natsTimer *t);
-
-// Returns the number of timers that have been created and not stopped.
-int
-nats_getTimersCount(void);
-
-// Returns the number of timers actually in the list. This should be
-// equal to nats_getTimersCount() or nats_getTimersCount() - 1 when a
-// timer thread is invoking a timer's callback.
-int
-nats_getTimersCountInList(void);
-
-natsStatus
-nats_postAsyncCbInfo(natsAsyncCbInfo *info);
-
 void
 nats_sslRegisterThreadForCleanup(void);
 
-natsStatus
-nats_sslInit(void);
-
-natsStatus
-natsLib_msgDeliveryPostControlMsg(natsSubscription *sub);
-
-natsStatus
-natsLib_msgDeliveryAssignWorker(natsSubscription *sub);
-
-bool
-natsLib_isLibHandlingMsgDeliveryByDefault(void);
-
-int64_t
-natsLib_defaultWriteDeadline(void);
-
-void
-natsLib_getMsgDeliveryPoolInfo(int *maxSize, int *size, int *idx, natsMsgDlvWorker ***workersArray);
-
 void
 nats_setNATSThreadKey(void);
 
@@ -936,4 +907,19 @@ jsSub_resetOrderedConsumer(natsSubscription *sub, uint64_t sseq);
 bool
 natsMsg_isJSCtrl(natsMsg *msg, int *ctrlType);
 
+static inline void nats_lockDispatcher(natsDispatcher *d)
+{
+    if (d->mu != NULL)
+        natsMutex_Lock(d->mu);
+}
+
+static inline void nats_unlockDispatcher(natsDispatcher *d)
+{
+    if (d->mu != NULL)
+        natsMutex_Unlock(d->mu);
+}
+
+void nats_dispatchThreadPool(void *arg);
+void nats_dispatchThreadOwn(void *arg);
+
 #endif /* NATSP_H_ */
