@@ -1705,7 +1705,7 @@ nats_Base32_Init(void)
     int         i;
 
     for (i=0; i<(int)sizeof(base32DecodeMap); i++)
-        base32DecodeMap[i] = (char) 0xFF;
+        base32DecodeMap[i] = '\xFF';
 
     for (i=0; i<alphaLen; i++)
         base32DecodeMap[(int)alphabet[i]] = (char) i;
@@ -1747,7 +1747,7 @@ nats_Base32_DecodeString(const char *src, char *dst, int dstMax, int *dstLen)
             dbuf[j] = base32DecodeMap[in];
             // If invalid character, report the position but as the number of character
             // since beginning, not array index.
-            if (dbuf[j] == (char) 0xFF)
+            if (dbuf[j] == '\xFF')
                 return nats_setError(NATS_ERR, "base32: invalid data at location %d", srcLen - remaining);
             j++;
         }
@@ -1799,7 +1799,7 @@ _base64Encode(const char *map, bool padding, const unsigned char *src, int srcLe
 
     *pDest = NULL;
 
-    if ((src == NULL) || (src[0] == '\0'))
+    if ((src == NULL) || (srcLen == 0))
         return NATS_OK;
 
     n = srcLen;
@@ -2064,10 +2064,9 @@ nats_HostIsIP(const char *host)
 }
 
 static bool
-_isLineAnHeader(const char *ptr)
+_isLineAnHeader(const char *ptr, int len)
 {
     char    *last   = NULL;
-    int     len     = 0;
     int     count   = 0;
     bool    done    = false;
 
@@ -2075,7 +2074,6 @@ _isLineAnHeader(const char *ptr)
     // the strict requirement is that it ends with at least 3 consecutive
     // `-` characters. It must also have 3 consecutive `-` before that.
     // So the minimum size would be 6.
-    len = (int) strlen(ptr);
     if (len < 6)
         return false;
 
@@ -2118,61 +2116,72 @@ _isLineAnHeader(const char *ptr)
     return false;
 }
 
+// Finds the next entire line in next, sets start to the beginning of the line,
+// updates next to point to the remaining bytes, and returns the line's length.
+// If there are no more non-empty lines, returns 0.
+static inline int
+_scan_line(const char **start, const char **next)
+{
+    const char  *ptr    = *next;
+    int         n       = 0;
+
+    // skip empty lines in the beginning
+    while (*ptr == '\r' || *ptr == '\n')
+        ptr++;
+    *start = ptr;
+
+    // Consume until we reach the end of the line
+    while (*ptr != '\r' && *ptr != '\n' && *ptr != '\0')
+    {
+        ptr++;
+        n++;
+    }
+
+    if (n == 0)
+        return 0;
+
+    // skip the subsequent empty lines, 'cause why not?
+    while (*ptr == '\r' || *ptr == '\n')
+        ptr++;
+    *next = ptr;
+
+    return n;
+}
+
 natsStatus
 nats_GetJWTOrSeed(char **val, const char *content, int item)
 {
-    natsStatus  s       = NATS_OK;
-    char        *pch    = NULL;
-    char        *str    = NULL;
-    char        *saved  = NULL;
-    int         curItem = 0;
-    int         orgLen  = 0;
-    char        *nt     = NULL;
+    natsStatus  s           = NATS_OK;
+    const char  *next       = content;
+    const char  *line       = NULL;
+    int         lineLen     = 0;
+    int         curItem     = 0;
+    const char  *saved      = NULL;
+    int         savedLen    = 0;
 
-    // First, make a copy of the original content since
-    // we are going to call strtok on it, which alters it.
-    str = NATS_STRDUP(content);
-    if (str == NULL)
-        return nats_setDefaultError(NATS_NO_MEMORY);
-
-    orgLen = (int) strlen(str);
-
-    pch = nats_strtok(str, "\n", &nt);
-    while (pch != NULL)
+    while ((lineLen = _scan_line(&line, &next)) > 0)
     {
-        if (_isLineAnHeader(pch))
+        if (_isLineAnHeader(line, lineLen))
         {
             // We got the start of the section. Save the next line
             // as the possible returned value if the following line
             // is a header too.
-            pch = nats_strtok(NULL, "\n", &nt);
-            saved = pch;
+            savedLen = _scan_line(&saved, &next);
+            if (savedLen == 0)
+                break; // premature end of file?
 
-            while (pch != NULL)
-            {
-                pch = nats_strtok(NULL, "\n", &nt);
-                if (pch == NULL)
-                    break;
-
-                // We tolerate empty string(s).
-                if (*pch == '\0')
-                    continue;
-
-                break;
-            }
-            if (pch == NULL)
-                break;
-
-            if (_isLineAnHeader(pch))
+            lineLen = _scan_line(&line, &next);
+            if (_isLineAnHeader(line, lineLen))
             {
                 // Is this the item we were looking for?
                 if (curItem == item)
                 {
                     // Return a copy of the saved line
-                    *val = NATS_STRDUP(saved);
+                    *val = NATS_CALLOC(savedLen + 1, 1);
                     if (*val == NULL)
                         s = nats_setDefaultError(NATS_NO_MEMORY);
-
+                    else
+                        memcpy(*val, saved, savedLen);
                     break;
                 }
                 else if (++curItem > 1)
@@ -2181,12 +2190,8 @@ nats_GetJWTOrSeed(char **val, const char *content, int item)
                 }
             }
         }
-        pch = nats_strtok(NULL, "\n", &nt);
     }
 
-    memset(str, 0, orgLen);
-    NATS_FREE(str);
-
     // Nothing was found, return NATS_NOT_FOUND but don't set the stack error.
     if ((s == NATS_OK) && (*val == NULL))
         return NATS_NOT_FOUND;
@@ -2238,11 +2243,11 @@ fmt_frac(char *buf, int w, uint64_t v, int prec, int *nw, uint64_t *nv)
     // Omit trailing zeros up to and including decimal point.
     bool print = false;
     int i;
-    int digit;
+    char digit;
 
     for (i = 0; i < prec; i++)
     {
-        digit = v % 10;
+        digit = (char)(v % 10);
         print = print || digit != 0;
         if (print)
         {
@@ -2551,7 +2556,66 @@ nats_freeMetadata(natsMetadata *md)
     {
         NATS_FREE((char *)md->List[i]);
     }
-    NATS_FREE(md->List);
+    NATS_FREE((char*) md->List);
     md->List = NULL;
     md->Count = 0;
 }
+
+
+// allocates a sufficiently large buffer and formats the strings into it, as a
+// ["unencoded-string-0","unencoded-string-1",...]. For an empty array of
+// strings returns "[]".
+natsStatus nats_formatStringArray(char **out, const char **strings, int count)
+{
+    natsStatus s = NATS_OK;
+    natsBuffer buf;
+    int len = 0;
+    int  i;
+
+    len++; // For the '['
+    for (i = 0; i < count; i++)
+    {
+        len += 2; // For the quotes
+        if (i > 0)
+            len++; // For the ','
+        if (strings[i] == NULL)
+            len += (int)strlen("(null)");
+        else
+            len += (int)strlen(strings[i]);
+    }
+    len++; // For the ']'
+    len++; // For the '\0'
+
+    s = natsBuf_Init(&buf, len);
+
+    natsBuf_AppendByte(&buf, '[');
+    for (i = 0; (s == NATS_OK) && (i < count); i++)
+    {
+        if (i > 0)
+        {
+            IFOK(s, natsBuf_AppendByte(&buf, ','));
+        }
+        IFOK(s, natsBuf_AppendByte(&buf, '"'));
+        if (strings[i] == NULL)
+        {
+            IFOK(s, natsBuf_Append(&buf, "(null)", -1));
+        }
+        else
+        {
+            IFOK(s, natsBuf_Append(&buf, strings[i], -1));
+        }
+        IFOK(s, natsBuf_AppendByte(&buf, '"'));
+    }
+
+    IFOK(s, natsBuf_AppendByte(&buf, ']'));
+    IFOK(s, natsBuf_AppendByte(&buf, '\0'));
+    
+    if (s != NATS_OK)
+    {
+        natsBuf_Cleanup(&buf);
+        return s;
+    }
+
+    *out = natsBuf_Data(&buf);
+    return NATS_OK;
+}
